@@ -8,6 +8,7 @@ from functools import reduce
 from igi_gc_reader.reader.excel_address import ExcelAddress, col_index_to_letter, col_letter_to_index
 from igi_gc_reader.reader.excel_reader import IExcelReader, has_value
 from igi_gc_reader.reader.classification import GcFileClass
+from igi_gc_reader.utils import gc_logger
 
 
 @dataclass
@@ -38,11 +39,6 @@ class PagePhase(Enum):
     InPageData = 5
 
 
-# TODO - (MAYBE) *** Future note for classifier ***
-#        distinguishing characteristic of this file class seems to be 🤞 the repeating metadata
-#        blocks at the top of each page - looking for any of 4 specific headers usually seen in 
-#        left metadata column. This could be a useful test when implementing a better classifier!
-
 # TODO - (MAYBE) Consider changing to a pattern where 
 #        we have an IPageState abstract class and a subclass for each state:
 
@@ -51,6 +47,10 @@ class PagePhase(Enum):
 #   def process(row)
 
 # then just go through states for each row and call transition (and process if no transition)
+
+
+# See readme.md at top level dir for overview of page handling logic
+# --- ---------
 @dataclass
 class GcPageClassOneReader(IGcPageReader):
     _xl_reader: IExcelReader
@@ -58,7 +58,8 @@ class GcPageClassOneReader(IGcPageReader):
     expected_metadata_headers: ClassVar[Set[str]] = {'Company:', 'Well Name:', 'Depth:', 
                                                       'Sampling Point:'}
     page_start_col: ClassVar[str] = 'A'
-    derived_data_header: ClassVar[str] = 'Miscellaneous Ratios'  # use as an indicator to stop
+    post_data_headers: ClassVar[Set[str]] = {  # use as an indicator to stop
+        'Miscellaneous Ratios', 'Carbon Number Distribution', 'Parameter'}
     start_page_search_from_row: ClassVar[int] = 40
     max_cols: ClassVar[int] = 60  # look for headers in all columns up to this
     end_after_n_empty_rows: ClassVar[int] = 100
@@ -78,9 +79,9 @@ class GcPageClassOneReader(IGcPageReader):
         # 3 stopping conditions as of 05/07/2021:
         #   1) Hit >= n empty rows consecutively (after finding start of data pages)
         #      (n configured as ClassVar with {n_empty_cols_for_row})
-        #   2) Detected a derived data header (derived data incl at end). This is 
+        #   2) Detected a head for ignored data at end e.g. derived data. This is 
         #      a backup for in case there is not a large gap from end of last data page
-        #      (derived data header configured as ClassVar: {derived_data_header})
+        #      (configured as ClassVar: {post_data_headers})
         #   3) Fallback - set a max num rows just incase of a logical error or unexpected
         #      file format to prevent possible infinite loop.
         #      (max rows configured as ClassVar: {max_rows})
@@ -89,13 +90,14 @@ class GcPageClassOneReader(IGcPageReader):
             consecutive_empty_rows = consecutive_empty_rows + 1 if is_empty_row else 0
 
             if consecutive_empty_rows >= self.end_after_n_empty_rows and found_start:
+                self.add_data_rows(current_page, data_row_col_pair_to_value)
                 if current_page.has_data():
                     yield current_page
                 return  # end cond 1: n consecutive empty rows
 
             # phase 1
             if phase == PagePhase.Seeking:
-                if self.is_start_of_new_page(row):
+                if not is_empty_row and self.is_start_of_new_page(row):
                     found_start = True
                     phase = PagePhase.InPageMetadata
 
@@ -112,10 +114,13 @@ class GcPageClassOneReader(IGcPageReader):
 
             # phase 4
             if phase == PagePhase.InPageHeaders:
-                if header_row_num <3:
+                if header_row_num <= 2:
                     for col, val in self.get_col_value_pairs(row):
+                        # handle header row 2 slight misalignment if needed...
+                        col = self.adj_col_for_off_by_ones(header_row_num, 
+                                                           header_col_to_values, col, val)
                         header_col_to_values[col].append(str(val))
-                        if val == self.derived_data_header:
+                        if val in self.post_data_headers:
                             return  # end cond 2: into derived data (last data page already returned)
                     header_row_num += 1  
                 else:
@@ -133,7 +138,9 @@ class GcPageClassOneReader(IGcPageReader):
             # phase 5
             if phase == PagePhase.InPageData:
                 is_new_page = self.is_start_of_new_page(row) 
-                if is_new_page or is_empty_row:
+                if is_empty_row or self.is_subheading(row):
+                    continue
+                if is_new_page:
                     if is_new_page:
                         phase = PagePhase.InPageMetadata
                     else:
@@ -151,18 +158,31 @@ class GcPageClassOneReader(IGcPageReader):
                 for col, val in self.get_col_value_pairs(row):
                     data_row_col_pair_to_value[(row, col)] = val
 
-    def is_empty_row(self, row:int):
+    def is_empty_row(self, row: int):
         check_cols = [col_index_to_letter(i) for i in range(0, self.n_empty_cols_for_row)]
         return not any(has_value(self.read(col, row)) for col in check_cols)
 
     def read(self, col: str, row: int) -> Any:
         return self._xl_reader.read_cell(ExcelAddress(f"{col}{row}"))
 
-    def is_start_of_new_page(self, row) -> bool:
+    def is_start_of_new_page(self, row: int) -> bool:
         col_a_cleaned = str(self.read(col=self.page_start_col, row=row)).strip()
         return col_a_cleaned in self.expected_metadata_headers
 
+    def is_subheading(self, row: int) -> bool:
+        """
+        Example subheadings seen in GCMSMS file provided by Paul: 
+        `330.3®217.2: Internal Standard`. All examples seen only have a value in col A.
+        """        
+        col_value_pairs = self.get_col_value_pairs(row)
+        col_A_has_value = next(col_value_pairs)[0] == 'A'
+        any_other_values = any(col_value_pairs)  # any left after taking element above?
+        return col_A_has_value and not any_other_values
+
     def get_col_value_pairs(self, row: int) -> Iterator[Tuple[str, Any]]:
+        """
+        Return all none empty values in row as col, value pairs.
+        """
         for col_idx in range(self.max_cols):
             col = col_index_to_letter(col_idx)
             val = self.read(col, row)
@@ -185,6 +205,27 @@ class GcPageClassOneReader(IGcPageReader):
                 val = data_row_col_pair_to_value.get((r, c))
                 data_row.append(val)
             page.data_rows.append(data_row)
+
+    def adj_col_for_off_by_ones(self, header_row_num, header_col_to_values, col, val):
+        """
+        Row 2 headers can sometimes align visually, but use a column that is off by
+        one relative to the row above. If one header row 2 and the col with a value
+        is not in the row 1 cols, assign to one +/- 1 col if found.
+        """
+        # handle off by one alignment issues i.e. col is +/- 1 on 2nd header row
+        if header_row_num == 2:
+            if col not in header_col_to_values:
+                this_col_idx = col_letter_to_index(col)
+                col_dist_pairs = [(c, abs(this_col_idx - col_letter_to_index(c)))
+                                                  for c in header_col_to_values]
+                off_by_ones = list(filter(lambda p: p[1] == 1, col_dist_pairs))
+                if len(off_by_ones) == 1:
+                    col = off_by_ones[0][0]  # correct column to match from row1
+                if len(off_by_ones) > 1:
+                    gc_logger.warning(f"Row 2 header {val} in col {col} does not align "
+                         f"with row 1. Multiple row 1 matches were close: {off_by_ones}, "
+                         f"left unchanged as it is not clear which it should join with.")
+        return col
             
 
 def get_page_reader(xl_reader: IExcelReader, file_class: GcFileClass) -> IGcPageReader:
@@ -207,8 +248,10 @@ def merge_pages(pages: List[GcPageData]) -> GcPageData:
 
 
 def get_page_data(xl_reader: IExcelReader, file_class: GcFileClass) -> GcPageData:
+    gc_logger.info("  getting page data...")
     page_reader = get_page_reader(xl_reader, file_class)
     pages = list(page_reader.get_pages())
     if not pages:
         raise StopIteration("No page data found for file.")
+    gc_logger.info(f"  found {len(pages)} pages of GC data.")
     return merge_pages(pages)
